///////////////////////////////////////////////////////////////////////////////
// This source file is part of the LuaPlus source distribution and is Copyright
// 2002 by Joshua C. Jensen (jjensen@workspacewhiz.com).
//
// The latest version may be obtained from http://wwhiz.com/LuaPlus/.
//
// The code presented in this file may be freely used and modified for all
// non-commercial and commercial purposes.
///////////////////////////////////////////////////////////////////////////////
#pragma once

///////////////////////////////////////////////////////////////////////////////
// namespace LuaPlus
///////////////////////////////////////////////////////////////////////////////
namespace LuaPlus
{

class LuaCall;
struct LuaArgNil {};

struct LuaRun
{
	LuaRun(int numResults = -1) throw()
	{
		m_numResults = numResults;
	}

	int m_numResults;
};

/**
	Representation of a Lua object.
**/
class LUA_API LuaObject
{
public:
	/**
	**/
	LuaObject() throw() :
		m_state(NULL),
		m_next(NULL),
		m_prev(NULL)
	{
		// Bad to have this barebones constructor... but useful for containers.
	}

	/**
		Various constructors accepting different parameters.
	**/
	LuaObject::LuaObject(LuaState* state) throw();

	/**
		Various constructors accepting different parameters.
	**/
	LuaObject( LuaState* state, int stackIndex ) throw();

	/**
		Various constructors accepting different parameters.
	**/
	LuaObject( LuaState& state, int stackIndex ) throw();

	/**
		Various constructors accepting different parameters.
	**/
	LuaObject( LuaState* state, const NAMESPACE_LUA_PREFIX lua_TObject* obj ) throw();

	/**
		Copy constructor.
	**/
	LuaObject( const LuaStackObject& src ) throw();

	/**
		Copy constructor.
	**/
	LuaObject( const LuaObject& src ) throw();

	/**
		Assignment operator.
	**/
	const LuaObject& operator=( const LuaObject& src ) throw();

	/**
		Assignment operator.
	**/
	const LuaObject& operator=( const LuaStackObject& src ) throw();

	~LuaObject();

	/**
		Retrieves the LuaState object associated with this LuaObject.
	**/
	LuaState* GetState() const;
	NAMESPACE_LUA_PREFIX lua_State* GetCState() const;
	operator NAMESPACE_LUA_PREFIX lua_State*() const;
	operator LuaState*() const;

	bool operator==(const LuaObject& right) const;
	bool operator<(const LuaObject& right) const;

	const char* GetTypeName() const;

	int GetType() const;

	bool IsNil() const;
	bool IsTable() const;
	bool IsUserData() const;
	bool IsCFunction() const;
	bool IsNumber() const;
	bool IsString() const;
	bool IsWString() const;
	bool IsFunction() const;
	bool IsNone() const;
	bool IsLightUserData() const;
	bool IsBoolean() const;

	int GetInteger() const;
	float GetFloat() const;
	double GetDouble() const;
	float GetNumber() const;
	const char* GetString() const;
	const wchar_t* GetWString() const;
	int StrLen();
	NAMESPACE_LUA_PREFIX lua_CFunction GetCFunction() const;
	void* GetUserData();
	const void* GetLuaPointer();
	void* GetLightUserData() const;
	bool GetBoolean() const;

	LuaStackObject PushStack();

	LuaStackObject __cdecl PCall( const char* argTypes, ... );
	LuaStackObject __cdecl PCall( int nresults, const char* argTypes, ... );

	LuaObject GetMetaTable();
	void SetMetaTable( LuaObject& valueObj );

	int GetN();
	void SetN(int n);

	void Insert(int index, LuaObject& obj);
	void Remove(int index = -1);
	void Sort();

	int GetCount();
	int GetTableCount();

	LuaObject Clone();

	LuaObject CreateTable( const char* key, int narray = 0, int lnhash = 0 );
	LuaObject CreateTable(int key, int narray = 0, int lnhash = 0);
	LuaObject CreateTable(LuaObject& key, int narray = 0, int lnhash = 0);

	void SetNil(const char* key);
	void SetNil(int key);
	void SetNil(LuaObject& key);
	void SetBoolean(const char* key, bool value);
	void SetBoolean(int key, bool value);
	void SetBoolean(LuaObject& key, bool value);
	void SetNumber(const char* key, double value);
	void SetNumber(int key, double value);
	void SetNumber(LuaObject& key, double value);
	void SetString(const char* key, const char* value);
	void SetString(int key, const char* value);
	void SetString(LuaObject& key, const char* value);
	void SetWString(const char* key, const wchar_t* value);
	void SetWString(int key, const wchar_t* value);
	void SetWString(LuaObject& key, const wchar_t* value);
	void SetUserData(const char* key, void* value);
	void SetUserData(int key, void* value);
	void SetUserData(LuaObject& key, void* value);
	void SetLightUserData(const char* key, void* value);
	void SetLightUserData(int key, void* value);
	void SetLightUserData(LuaObject& key, void* value);
	void SetObject(const char* key, LuaObject& value);
	void SetObject(int key, LuaObject& value);
	void SetObject(LuaObject& key, LuaObject& value);

	void AssignNil();
	void AssignBoolean(bool value);
	void AssignNumber(double value);
	void AssignString(const char* value);
	void AssignWString(const wchar_t* value);
	void AssignUserData(void* value);
	void AssignLightUserData(void* value);
	void AssignObject(LuaObject& value);		// Should this function be removed??
	void AssignNewTable(int narray = 0, int numhash = 0);
	void AssignTObject(NAMESPACE_LUA_PREFIX lua_TObject* value);

	LuaObject GetByName( const char* name );
	LuaObject GetByIndex( int index );
	LuaObject GetByObject( const LuaStackObject& obj );
	LuaObject GetByObject( const LuaObject& obj );
	LuaObject operator[]( const char* name );
	LuaObject operator[]( int index );
	LuaObject operator[]( const LuaStackObject& obj );
	LuaObject operator[]( const LuaObject& obj );

	void Register( const char* funcName, NAMESPACE_LUA_PREFIX lua_CFunction function );
	void Register( const char* funcName, LuaStateCFunction function );
	void Register( const char* funcName, LSFunctor functor );
	void Register( const LuaFunctionList* functionList );
	void Register( const LuaFunctorList* functionList );
	void Unregister( const char* funcName );
	void Unregister( const LuaFunctionList* functionList );
	void Unregister( const LuaFunctorList* functionList );

	LuaCall operator()();
	LuaStackObject PCall(int nresults, const char* argTypes, va_list args);

	// Private functions and data
	void AddToUsedList(LuaObject* object);
	void RemoveFromUsedList(LuaObject* object);

	LuaState* m_state;
	LuaObject* m_next;		   // only valid when in free list
	LuaObject* m_prev;		   // only valid when in used list
#ifdef BUILDING_LUAPLUS
	lua_TObject m_object;
#else BUILDING_LUAPLUS
	unsigned char m_object[12];
#endif BUILDING_LUAPLUS

protected:
	void SetTableHelper(const char* key, NAMESPACE_LUA_PREFIX lua_TObject& valueObj);
	void SetTableHelper(int key, NAMESPACE_LUA_PREFIX lua_TObject& valueObj);
	void SetTableHelper(LuaObject& key, NAMESPACE_LUA_PREFIX lua_TObject& valueObj);
};


class LuaCall
{
public:
	LuaCall(LuaObject& functionObj);

	LuaCall& operator<<(const LuaArgNil& /*value*/);
	LuaCall& operator<<(float value);
	LuaCall& operator<<(double value);
	LuaCall& operator<<(int value);
	LuaCall& operator<<(const char* value);
	LuaCall& operator<<(const wchar_t* value);
	LuaCall& operator<<(NAMESPACE_LUA_PREFIX lua_CFunction value);
	LuaCall& operator<<(LuaStateCFunction value);
	LuaCall& operator<<(bool value);
	LuaCall& operator<<(void* value);
	LuaCall& operator<<(LuaStackObject& obj);
	LuaCall& operator<<(LuaObject& obj);
	LuaStackObject operator<<(const LuaRun& /*run*/);
	LuaCall& operator=(const LuaCall& src);

protected:
	LuaState* m_state;
	LuaObject m_functionObj;
	int m_numArgs;
	int m_startResults;
};


} // namespace LuaPlus

#include "LuaState.h"

