///////////////////////////////////////////////////////////////////////////////
// This source file is part of the LuaPlus source distribution and is Copyright
// 2002 by Joshua C. Jensen (jjensen@workspacewhiz.com).
// The latest version may be obtained from http://www.workspacewhiz.com/.
//
// The code presented in this file may be freely used and modified for all
// non-commercial and commercial purposes.
///////////////////////////////////////////////////////////////////////////////
#ifdef WIN32
#include <windows.h>
#endif WIN32

#include <io.h>
#include "LuaPlus.h"
#include "LuaPlusFunctions.h"
LUA_EXTERN_C_BEGIN
#include "include/lua.h"
#include "include/lauxlib.h"
LUA_EXTERN_C_END

USING_NAMESPACE_LUA

#pragma warning( disable: 4100 )

namespace LuaPlus {

#ifdef WIN32

struct FindFileInfo
{
	FindFileInfo() : m_handle(NULL) {}
	~FindFileInfo() {  CloseHandle( m_handle );  }
	HANDLE m_handle;
	WIN32_FIND_DATA m_fd;
};

#endif WIN32

static int LS_Sleep( LuaState* state, LuaStackObject* args )
{
	if (!args[1].IsNumber())
		return 0;

	double sleepValue = args[1].GetNumber();	sleepValue;
#ifdef WIN32
	Sleep((DWORD)sleepValue);
#endif WIN32
	return 0;
}

	
static int LS_FindFirstFile( LuaState* state, LuaStackObject* args )
{
	const char* wildcard = args[ 1 ].GetString();	wildcard;

#ifdef WIN32
	FindFileInfo* info = new FindFileInfo;
	info->m_handle = FindFirstFile(wildcard, &info->m_fd);
	if (info->m_handle == INVALID_HANDLE_VALUE)
	{
		delete info;
		return 0;
	}

	state->NewUserDataBox(info);
	state->GetGlobal("FileFind");
	state->SetMetaTable(-2);

	return 1;
#else !WIN32
	return 0;
#endif WIN32
}


static int LS_FindNextFile( LuaState* state, LuaStackObject* args )
{
	if ( !args[ 1 ].IsUserData())
		return 0;

#ifdef WIN32
	FindFileInfo* info = static_cast<FindFileInfo*>( args[ 1 ].GetUserData() );
	BOOL ret = FindNextFile(info->m_handle, &info->m_fd);
	if (!ret)
	{
		FindClose(info->m_handle);
		info->m_handle = NULL;
		return 0;
	}

	state->PushBoolean(true);
	return 1;
#else !WIN32
	return 0;
#endif WIN32
}


static int LS_FindClose( LuaState* state, LuaStackObject* args )
{
	if ( !args[ 1 ].IsUserData() )
		return 0;

#ifdef WIN32
	FindFileInfo* info = static_cast<FindFileInfo*>( args[ 1 ].GetUserData() );
	delete info;
#endif WIN32
	return 0;
}


static int LS_FindGetFileName( LuaState* state, LuaStackObject* args )
{
	if ( !args[ 1 ].IsUserData() )
		return 0;

#ifdef WIN32
	FindFileInfo* info = static_cast<FindFileInfo*>( args[ 1 ].GetUserData() );
	state->PushString(info->m_fd.cFileName);
#endif WIN32

	return 1;
}


static int LS_FindIsDirectory( LuaState* state, LuaStackObject* args )
{
	if ( !args[ 1 ].IsUserData() )
		return 0;

#ifdef WIN32
	FindFileInfo* info = static_cast<FindFileInfo*>( args[ 1 ].GetUserData() );
	state->PushBoolean((info->m_fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0);

	return 1;
#else !WIN32
	return 0;
#endif WIN32
}

static int LS_GetFileSize( LuaState* state, LuaStackObject* args )
{
	if ( !args[ 1 ].IsString())
		return 0;

#ifdef WIN32
	WIN32_FIND_DATA fd;
	HANDLE handle = FindFirstFile( args[ 1 ].GetString(), &fd );
	state->PushNumber(fd.nFileSizeLow);
	FindClose(handle);
	return 1;
#else !WIN32
	return 0;
#endif WIN32

}

static int LS_GetFileWriteTime( LuaState* state, LuaStackObject* args )
{
	LuaStackObject fileNameObj(state, 1);
	if (!fileNameObj.IsString())
		return 0;

#ifdef WIN32
	WIN32_FIND_DATA fd;
	HANDLE handle = FindFirstFile(fileNameObj.GetString(), &fd);
	state->PushNumber(fd.ftLastWriteTime.dwLowDateTime);
	state->PushNumber(fd.ftLastWriteTime.dwHighDateTime);
	FindClose(handle);
	return 2;
#else !WIN32
	return 0;
#endif WIN32
}

static int LS_SetFileWriteTime( LuaState* state, LuaStackObject* args )
{
	if ( !args[ 1 ].IsString() )
		return 0;

#ifdef WIN32
	FILETIME createTime;
	FILETIME accessTime;
	FILETIME writeTime;
	LPCSTR fileName = args[ 1 ].GetString();
	HANDLE handle = CreateFile(fileName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	GetFileTime(handle, &createTime, &accessTime, &writeTime);
	writeTime.dwLowDateTime = args[ 2 ].GetInteger();
	writeTime.dwHighDateTime = args[ 3 ].GetInteger();
	SetFileTime(handle, &createTime, &accessTime, &writeTime);
	CloseHandle(handle);
	return 0;
#else !WIN32
	return 0;
#endif WIN32
}

static int LS_access( LuaState* state, LuaStackObject* args )
{
	LuaStackObject& fileNameObj = args[1];
	if (!fileNameObj.IsString())
		return 0;
	LuaStackObject& typeObj = args[2];
	const char* type = NULL;
	if (typeObj.IsString())
		type = typeObj.GetString();

	int access = 0;
	if (type)
	{
		if (strcmp(type, "w") == 0)
			access = 2;
		else if (strcmp(type, "r") == 0)
			access = 4;
		else if (strcmp(type, "rw") == 0)
			access = 6;
	}

	state->PushNumber(_access(fileNameObj.GetString(), access));
	return 1;
}



static int LS_NewTableSize( LuaState* state, LuaStackObject* args )
{
//jj	lua_newtablesize(state, (int)lua_tonumber(state, 1), (int)lua_tonumber(state, 2));
	lua_newtable(*state);
	return 1;
}


typedef void (*fnLuaPlus_RegisterModule)( LuaState* state );
typedef void (*fnLuaPlus_UnregisterModule)( LuaState* state );


static int LS_import( LuaState* state, LuaStackObject* args )
{
#ifndef LUAPLUS_LIB
	LPCSTR moduleName = args[ 1 ].GetString();

#ifdef _DEBUG
	const char* luaplusdllName = "LuaPlusD.dll";
#else _DEBUG
	const char* luaplusdllName = "LuaPlus.dll";
#endif _DEBUG

	char path[ _MAX_PATH ];
	GetModuleFileName(GetModuleHandle(luaplusdllName), path, _MAX_PATH);
	char* slashPtr = strrchr( path, '\\' );
	slashPtr++;
	*slashPtr = 0;
	strcat( path, "modules/" );
	strcat( path, moduleName );
	strcat( path, ".dll" );
#ifdef _DEBUG
	strcat( path, "d" );
#endif _DEBUG
	strlwr( path );

	LuaObject importModulesObj = state->GetRegistry()[ "ImportModules" ];
	if ( importModulesObj.IsNil() )
	{
		importModulesObj = state->GetRegistry().CreateTable( "ImportModules" );
	}

	LuaObject moduleObj = importModulesObj[ path ];
	if ( !moduleObj.IsNil() )
	{
		// Already exists.
		state->PushBoolean( true );
		return 1;
	}

	HINSTANCE hInst = LoadLibrary( path );
	if ( hInst == INVALID_HANDLE_VALUE )
	{
		// Didn't load successfully.
		state->PushBoolean( false );
		return 1;
	}

	fnLuaPlus_RegisterModule registerModule = (fnLuaPlus_RegisterModule)
			GetProcAddress( hInst , "LuaPlus_RegisterModule" );
	if ( !registerModule )
	{
		// Isn't a LuaPlus DLL.
		state->PushBoolean( false );
		return 1;
	}

	(*registerModule)( state );

	state->PushBoolean( true );
#else !LUAPLUS_LIB
	state->PushBoolean(false);
#endif LUAPLUS_LIB
	return 1;
}
	
static const LuaFunctionList s_fileFindFunctions[] =
{
	{ "First", LS_FindFirstFile },
	{ "Next", LS_FindNextFile },
	{ "Close", LS_FindClose },
	{ "GetFileName", LS_FindGetFileName },
	{ "IsDirectory", LS_FindIsDirectory },
	{ NULL, NULL },
};


static const LuaFunctionList s_fileFunctions[] =
{
	{ "GetSize",		LS_GetFileSize },
	{ "access",			LS_access },
	{ NULL, NULL },
};


void ScriptFunctionsRegister( LuaState* state )
{
	LuaObject globalsObj = state->GetGlobals();
	globalsObj.Register("FindFirstFile",	LS_FindFirstFile);
	globalsObj.Register("FindNextFile",		LS_FindNextFile);
	globalsObj.Register("FindClose",		LS_FindClose);
	globalsObj.Register("FindGetFileName",	LS_FindGetFileName);
	globalsObj.Register("FindIsDirectory",	LS_FindIsDirectory);
	globalsObj.Register("GetFileSize",		LS_GetFileSize);
	globalsObj.Register("newtablesize",		LS_NewTableSize);
	globalsObj.Register("access",			LS_access);
	globalsObj.Register( "import",			LS_import );
	LuaObject fileObj = state->GetRegistry()["FileHandle"];
	fileObj.Register(s_fileFunctions);
	
	LuaObject fileFindObj = globalsObj.CreateTable("FileFind");
	fileFindObj.Register(s_fileFindFunctions);

	LuaObject ioObj = globalsObj["io"];
	ioObj.Register( "GetWriteTime",	LS_GetFileWriteTime );
	ioObj.Register( "SetWriteTime",	LS_SetFileWriteTime );

	LuaObject osObj = globalsObj["os"];
	osObj.Register("Sleep",	LS_Sleep);
}

} // namespace LuaPlus





/*
--[[function io.lines(file, openType)
	if type( file ) == "string" then
		if file:sub( 1, 1 ) == "|" then
			file = io.popen( file:sub( 2 ), openType )
		else
			file = io.open( file, openType )
		end
	end
	
	return function ()
		return file:read() or (assert(file:close()) and nil), i
	end
end


function FullLookup(table, variableName)
	local lastPos = 1
	while 1 do
		local curPos = strfind(variableName, "%.", lastPos)
		if not curPos then
			local var = strsub(variableName, lastPos)
			local num = tonumber(var)
			if num then
				var = num
			end
			return table[var]
		end

		local var = strsub(variableName, lastPos, curPos - 1)
		local num = tonumber(var)
		if num then
			var = num
		end
		lastPos = curPos + 1
		table = table[var]
		if not table then
			return nil
		end
	end
end


function DirProcessFiles(path, wildcard, func)
	if type(func) ~= "function" then
		return
	end

	path = path:gsub("/", "\\")
	if path:sub(path:len()) ~= '\\' then
		path = path .. "\\"
	end

	local handle = FileFind.First(path .. wildcard)

	while 1 do
		local fileName = handle:GetFileName()
		local isDirectory = handle:IsDirectory()
		func(path, fileName)
		if not handle:Next() then
			break
		end
	end

	handle:Close()
end


function DirProcessFilesRecursive(path, func, ext)
	if ext then
		ext = ext:lower()
	end

	if type(func) ~= "function" then
		return
	end

	path = path:gsub("/", "\\")
	if path:sub(path:len()) ~= '\\' then
		path = path .. "\\"
	end

	function ReadDirectory(path)
		local handle = FileFind.First(path .. "*.*")

		while handle do
			local fileName = handle:GetFileName()
			local isDirectory = handle:IsDirectory()
			if isDirectory then
				if fileName ~= "." and fileName ~= ".." then
					ReadDirectory(path .. fileName .. "\\")
				end
			else
				if ext then
					local dotPos, endPos, foundExt = fileName:find("%.(%w*)$")
					if foundExt then
						foundExt = foundExt:lower()
						if ext == foundExt then
							func(path, fileName)
						end
					end
				else
					func(path, fileName)
				end
			end
			if not handle:Next() then
				break
			end
		end
		
		if handle then
			handle:Close()
		end
	end

	ReadDirectory(path)
end
]]
*/