///////////////////////////////////////////////////////////////////////////////
// This source file is part of the LuaPlus source distribution and is Copyright
// 2002 by Joshua C. Jensen (jjensen@workspacewhiz.com).
//
// The latest version may be obtained from http://wwhiz.com/LuaPlus/.
//
// The code presented in this file may be freely used and modified for all
// non-commercial and commercial purposes.
///////////////////////////////////////////////////////////////////////////////
#pragma once

#include "LuaObject.h"

///////////////////////////////////////////////////////////////////////////////
// namespace LuaPlus
///////////////////////////////////////////////////////////////////////////////
namespace LuaPlus
{

/**
	A lightweight lua_State wrapper.
**/
class LUA_API LuaState
{
public:
	enum ErrorCodes
	{
		/* error codes for `lua_load' and `lua_pcall' */
		ERRRUN = 1,
		ERRFILE = 2,
		ERRSYNTAX = 3,
		ERRMEM = 4,
		ERRERR = 5,
	};

	enum Types
	{
		TNONE = (-1),
		TNIL	= 0,
		TBOOLEAN	= 1,
		TLIGHTUSERDATA	= 2,
		TNUMBER = 3,
		TSTRING = 4,
		TTABLE = 5,
		TFUNCTION = 6,
		TUSERDATA = 7,
		TTHREAD = 8,
		TWSTRING = 9,
		NTYPES = 11,
	};


	///////////////////////////////////////////////////////////////////////////
	static LuaState* Create(bool initStandardLibrary = true, bool multithreaded = false);
	static LuaState* CreateThread(LuaState* parentState);
	static LuaState* CastState(NAMESPACE_LUA_PREFIX lua_State* state);
	static void Destroy(LuaState* state);

	NAMESPACE_LUA_PREFIX lua_CFunction AtPanic( NAMESPACE_LUA_PREFIX lua_CFunction panicf );

	// LuaObject
	LuaObject GetGlobals() throw();
	LuaObject GetGlobal( const char *name );
	LuaObject GetRegistry();

	int Equal(const LuaObject& o1, const LuaObject& o2);
	int LessThan(const LuaObject& o1, const LuaObject& o2);

	// Stack functions.
	LuaStackObject Stack( int index );
	int GetTop();
	void SetTop( int index );
	void PushValue( int index );
	void PushValue( LuaStackObject object );
	void RemoveStack( int index );
	void InsertStack( int index );
	void ReplaceStack( int index );
	int CheckStack(int size);
	void XMove(LuaState* to, int n);
	int Equal_Stack( int index1, int index2 );
	int LessThan_Stack( int index1, int index2 );

	// push functions (C -> stack)
	LuaStackObject PushNil();
	LuaStackObject PushNumber( double n );
	LuaStackObject PushLString( const char *s, size_t len );
	LuaStackObject PushLWString( const wchar_t* s, size_t len );
	LuaStackObject PushString( const char *s );
	LuaStackObject PushWString( const wchar_t* s );
	LuaStackObject PushCClosure( NAMESPACE_LUA_PREFIX lua_CFunction fn, int n );
	LuaStackObject PushCClosure( LuaStateCFunction fn, int n );
	LuaStackObject PushCFunction( NAMESPACE_LUA_PREFIX lua_CFunction f );
	LuaStackObject PushCFunction( LuaStateCFunction f );
	LuaStackObject PushBoolean( bool value );
	LuaStackObject PushLightUserData( void* p );

	// get functions (Lua -> stack)
	void GetTable( int index );
	void RawGet( int index );
	void RawGetI( int index, int n );
	LuaStackObject NewTable( int size = 0 );
	LuaStackObject GetMetaTable( int index );
	LuaStackObject GetDefaultMetaTable( int type );
	LuaStackObject GetGlobals_Stack();
	LuaStackObject GetGlobal_Stack( const char *name );
	LuaStackObject GetRegistry_Stack();
	LuaStackObject GetRef( int ref );


	// set functions(stack -> Lua)
	void SetTable( int index );
	void RawSet( int index );
	void RawSetI( int index, int n );
	void SetMetaTable( int index );
	void SetDefaultMetaTable( int type );
	void SetGlobals( int index );
	void SetGlobal( const char *name );
	int Ref( int lock );
	void Unref( int ref );

	// `load' and `do' functions (load and run Lua code)
	void Call( int nargs, int nresults );
	int PCall( int nargs, int nresults, int errf );
	int CPCall(NAMESPACE_LUA_PREFIX lua_CFunction func, void* ud);

	int Dump(lua_Chunkwriter writer, void* data);

	int LoadFile( const char* filename );
	int DoFile( const char *filename );
	int DoString( const char *str );
	int LoadBuffer( const char* buff, size_t size, const char* name );
	int DoBuffer( const char *buff, size_t size, const char *name );

	// Coroutine functions
	int CoYield( int nresults );
	int CoResume( int narg );

	// Garbage-collection functions
	int GetGCThreshold();
	int GetGCCount();
	void SetGCThreshold( int newthreshold );

	// Miscellaneous functions
	int Error();
    int Error(const char * msg);

	int Next( int index );
	int GetN( int index );

	void Concat( int n );

	void* NewUserData( size_t size );
	LuaObject NewUserDataBox( void* u );
	LuaStackObject NewUserDataBox_Stack( void* u );

	// Helper functions
	void Pop();
	void Pop( int amount );

	void CollectGarbage();

	bool DumpFile(const char* filename, const char* name, LuaObject& value,
						bool alphabetical = true, int indentLevel = 0,
						unsigned int maxIndentLevel = 0xffffffff,
						bool writeAll = false, bool writeTablePointers = false);
	bool DumpObject(LuaStateOutFile& file, const char* name, LuaObject& value,
						bool alphabetical = true, int indentLevel = 0,
						unsigned int maxIndentLevel = 0xffffffff,
						bool writeAll = false, bool writeTablePointers = false);
	bool DumpGlobals(const char* filename, bool alphabetical = true,
						unsigned int maxIndentLevel = 0xFFFFFFFF,
						bool writeAll = false, bool writeTablePointers = false);

	operator NAMESPACE_LUA_PREFIX lua_State*()						{  return m_state;  }
	NAMESPACE_LUA_PREFIX lua_State* GetCState()						{  return m_state;  }

	// Debug functions.
	int GetStack( int level, NAMESPACE_LUA_PREFIX lua_Debug* ar );
	int GetInfo( const char* what, NAMESPACE_LUA_PREFIX lua_Debug* ar );
	const char* GetLocal( const NAMESPACE_LUA_PREFIX lua_Debug* ar, int n );
	const char* SetLocal( const NAMESPACE_LUA_PREFIX lua_Debug* ar, int n );

	int SetHook( NAMESPACE_LUA_PREFIX lua_Hook func, int mask, int count );
	NAMESPACE_LUA_PREFIX lua_Hook GetHook();
	unsigned long GetHookMask();

	// Extra
	LuaStackObject BoxPointer(void* u);
	void* UnBoxPointer(int stackIndex);

	NAMESPACE_LUA_PREFIX lua_State* m_state;
	bool m_ownState;
	LuaStackObject m_tweakMeToSeeStuff;
	LuaStackObject m_callbackArgs[20];

	// We want the storage space, but we can't have the constructor run.
	LuaObject* GetHeadObject()			{  return (LuaObject*)m_headObject;  }
	LuaObject* GetTailObject()			{  return (LuaObject*)m_tailObject;  }

protected:
	LuaState( bool initStandardLibrary = true, bool multithreaded = false );
	LuaState( LuaState* script, bool initStandardLibrary = true );
	LuaState( NAMESPACE_LUA_PREFIX lua_State* state );
	~LuaState();
	LuaState& operator=(LuaState& src);		// Not implemented.

	void Init( bool initStandardLibrary );

	bool CallFormatting( LuaObject& tableObj, LuaStateOutFile& file, int indentLevel,
			bool writeAll, bool alphabetical, bool writeTablePointers,
			unsigned int maxIndentLevel );

	void SetupStateEx();

	unsigned char m_headObject[sizeof(LuaObject)];
	unsigned char m_tailObject[sizeof(LuaObject)];
	LuaObject m_threadObj;
};


class LuaStateAuto
{
public:
    operator LuaState*()							{  return m_state;  }
    operator const LuaState*() const				{  return m_state;  }
    operator LuaState*() const						{  return m_state;  }
    LuaState& operator*()							{  return *m_state; }
    const LuaState& operator*() const				{  return *m_state; }
    LuaState* operator->()							{  return m_state;  }
    const LuaState* operator->() const				{  return m_state;  }
	LuaState* Get() const							{  return m_state;  }

	LuaStateAuto() : m_state(NULL) {}
    LuaStateAuto(LuaState* newState) : m_state(newState) {}
	LuaStateAuto& operator=(LuaState* newState)
	{
		Assign(newState);
		return *this;
	}

	~LuaStateAuto()
	{
		Assign(NULL);
	}

    LuaState* m_state;

protected:
    LuaStateAuto(const LuaStateAuto&);					// Not implemented.
    LuaStateAuto& operator=(const LuaStateAuto&);		// Not implemented.

	void Assign(LuaState* state)
	{
		if (m_state)
			LuaState::Destroy(m_state);
		m_state = state;
	}
};


class LuaStateOwner : public LuaStateAuto
{
public:
    LuaStateOwner(bool initStandardLibrary = true)
	{
		m_state = LuaState::Create(initStandardLibrary);
	}

    LuaStateOwner(LuaState* newState) : LuaStateAuto(newState) {}
	LuaStateOwner& operator=(LuaState* newState)
	{
		Assign(newState);
		return *this;
	}

	~LuaStateOwner()
	{
		Assign(NULL);
	}

private:
    LuaStateOwner(const LuaStateOwner&);				// Not implemented.
    LuaStateOwner& operator=(const LuaStateOwner&);		// Not implemented.
};


} // namespace LuaPlus

#ifdef LUAPLUS_ENABLE_INLINES
#include "LuaState.inl"
#endif LUAPLUS_ENABLE_INLINES

